// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#ifndef DUNE_ASSEMBLER_ARCHETYPES_LOCALBOUNDARYINTERSECTIONMATRIXASSEMBLER_HH
#define DUNE_ASSEMBLER_ARCHETYPES_LOCALBOUNDARYINTERSECTIONMATRIXASSEMBLER_HH

#include <dune/assembler/concepts.hh>
#include <dune/assembler/localviewpair.hh>

namespace Dune::Assembler::Concepts::Archetypes {

/**
 * Local assembler of a local vector on a boundary intersection.
 *
 * \tparam LocalView   Type of the local view associated to a function space basis.
 *
 * \b Requirements:
 * - `LocalView` is a model of Concepts::LocalView,
 */
template <Concepts::LocalView LocalView>
class LocalBoundaryIntersectionVectorAssembler
{
public:
  using GridView = typename LocalView::GridView;
  using Element = typename GridView::template Codim<0>::Entity;
  using Intersection = typename GridView::Intersection;

public:
  /**
   * This function is called typically before the elements of the grid are traversed.
   * Thus, the local views are not necessarily bound to any element.
   * The lifetime of the passed local view is longer than the lifetime of the local assembler.
   *
   * \param lv   A local views associated to a function space basis
   */
  void bindLocalView (const LocalView& lv) noexcept;

  /**
   * When this function is called, the local view is already bound to this element.
   * The lifetime of the passed element is longer than the lifetime of the local assembler.
   *
   * \param e  Grid element the local assembler can be bound to.
   * \note This function is called for all boundary elements. It is not called for inner elements
   *       unless this assembler is also an element or intersection assembler.
   * \note The user might check `e.hasBoundaryIntersections()`
   *
   * [[pre: localView.bound()]]
   */
  void bindElement (const Element& e) noexcept;

  /**
   * The vector `v` is assumed to be large enough to be indexable by the local view indices. It is not
   * guaranteed that the vector has the exact size as the local view.
   * It is not allowed to resize the vector.
   * The vector is already initialized and the assembler must only add its contributions to
   * the vector and not replace values.
   *
   * \param is  An intersection to assembler the local vector on.
   * \param v   A vector to be filled with the local element contributions.
   * \note The local assembler implementation should allow flexible matrix types as argument.
   * \note This function is called for all domain boundary intersections. It is not called
   *       for other inner or processor boundary intersections unless this assembler is also
   *       an intersection assembler.
   * \note The user might check `is.boundary()`
   *
   * [[pre: v.size() == testLocalView.size()]]
   * [[pre: bindElement was called before with current element]]
   */
  template <Concepts::Vector LocalVector>
  void assembleBoundaryIntersectionVector (const Intersection& is, LocalVector& v) noexcept;
};

} // namespace Dune::Assembler::Concepts::Archetypes

#endif // DUNE_ASSEMBLER_ARCHETYPES_LOCALBOUNDARYINTERSECTIONMATRIXASSEMBLER_HH
