// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#ifndef DUNE_ASSEMBLER_ARCHETYPES_LOCALELEMENTMATRIXASSEMBLER_HH
#define DUNE_ASSEMBLER_ARCHETYPES_LOCALELEMENTMATRIXASSEMBLER_HH

#include <dune/assembler/concepts.hh>

namespace Dune::Assembler::Concepts::Archetypes {

/**
 * Local assembler of a linear Matrix on an element.
 *
 * \tparam RowLocalView   Type of the local view associated to the test space.
 * \tparam ColLocalView  Type of the local view associated to the trial space.
 *
 * \b Requirements:
 * - `RowLocalView` is a model of Concepts::LocalView,
 * - `ColLocalView` is a model of Concepts::LocalView.
 */
template <Concepts::LocalView RowLocalView,
          Concepts::LocalView ColLocalView>
class LocalElementMatrixAssembler
{
public:
  using GridView = typename RowLocalView::GridView;
  using Element = typename GridView::template Codim<0>::Entity;

public:
  /**
   * This function is called typically before the elements of the grid are traversed.
   * Thus, the local views are not necessarily bound to any element.
   * The lifetime of the passed local view is longer than the lifetime of the local assembler.
   *
   * \param rlv  A local view of the test space
   * \param clv  A local view of the trial space
   */
  void bindLocalViews (const RowLocalView& rlv, const ColLocalView& clv) noexcept;

  /**
   * When this function is called, the test and trial local views are already bound to this element.
   * The lifetime of the passed element is longer than the lifetime of the local assembler.
   *
   * \param e  Grid element the local assembler can be bound to.
   *
   * [[pre: rowLocalView.bound() && colLocalView.bound()]]
   */
  void bindElement (const Element& e) noexcept;

  /**
   * The matrix `m` is assumed to be large enough to be indexable by all local view indices.
   * It is not guaranteed that the matrix has the exact size as the local views.
   * It is not allowed to resize the matrix.
   * The matrix is initialized and the assembler must only add its contributions to
   * the matrix and not replace values.
   *
   * \param m  A matrix to be filled with the local element contributions.
   * \param e  Grid element the local assembler is already bound to.
   * \note The local assembler implementation should allow flexible matrix types as argument.
   *
   * [[pre: m.rows() == rowLocalView.size() && m.cols() == colLocalView.size()]]
   * [[pre: bindElement was called with current element]]
   */
  template <Concepts::Matrix LocalMatrix>
  void assembleElementMatrix (const Element& e, LocalMatrix& m) noexcept;

  /**
   * Similar as assembleElementMatrix but for a local pattern instead of a matrix.
   * The local pattern can be accessed by `p.add(row,col)`. Additionally, it is possible to
   * set a local pattern to be full, by `p.addAll()`.
   */
  template <Concepts::LocalMatrixPattern LocalPattern>
  void assembleElementMatrixPattern (const Element& e, LocalPattern& p) noexcept;

};

} // namespace Dune::Assembler::Concepts::Archetypes

#endif // #define DUNE_ASSEMBLER_ARCHETYPES_LOCALELEMENTMATRIXASSEMBLER_HH
