// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#ifndef DUNE_ASSEMBLER_ARCHETYPES_LOCALINTERIORINTERSECTIONMATRIXASSEMBLER_HH
#define DUNE_ASSEMBLER_ARCHETYPES_LOCALINTERIORINTERSECTIONMATRIXASSEMBLER_HH

#include <dune/assembler/concepts.hh>

namespace Dune::Assembler::Concepts::Archetypes {

/**
 * Local assembler of a linear Matrix on an interior intersection.
 *
 * \tparam RowLocalView   Type of the local view associated to the test space.
 * \tparam ColLocalView  Type of the local view associated to the trial space.
 *
 * \b Requirements:
 * - `RowLocalView` is a model of Concepts::LocalView,
 * - `ColLocalView` is a model of Concepts::LocalView.
 */
template <Concepts::LocalView RowLocalView,
          Concepts::LocalView ColLocalView>
class LocalInteriorIntersectionMatrixAssembler
{
public:
  using GridView = typename RowLocalView::GridView;
  using Element = typename GridView::template Codim<0>::Entity;
  using Intersection = typename GridView::Intersection;

public:
  /**
   * This function is called typically before the elements of the grid are traversed.
   * Thus, the local views are not necessarily bound to any element.
   * The lifetime of the passed local view is longer than the lifetime of the local assembler.
   *
   * \param rlv  A local view of the test space
   * \param clv  A local view of the trial space
   */
  void bindLocalViews (const RowLocalView& rlv, const ColLocalView& clv) noexcept;

  /**
   * This function is called typically before the elements of the grid are traversed.
   * Thus, the local views are not necessarily bound to any element.
   * The lifetime of the passed local view is longer than the lifetime of the local assembler.
   *
   * \param rlv  A local view of the test space that will be bound to an intersection-outside element.
   * \param clv  A local view of the trial space that will be bound to an intersection-outside element.
   */
  void bindOutsideLocalViews (const RowLocalView& rlv, const ColLocalView& clv) noexcept;

  /**
   * When this function is called, the test and trial local views are already bound to this element.
   * The lifetime of the passed element is longer than the lifetime of the local assembler.
   *
   * \param e  Grid element the local assembler can be bound to.
   * \note The user might check `e.hasBoundaryIntersections()`
   *
   * [[pre: rowLocalView.bound() && colLocalView.bound()]]
   */
  void bindElement (const Element& e) noexcept;

  /**
   * The matrix `m` is a 2x2 block-matrix for `[ii, io; oi, oo]` couplings with `i...inside`
   * and `o...outside. Each submatrix `m[i][j]` is assumed to be large enough to be indexable
   * by all corresponding inside/outside local view indices.
   * It is not guaranteed that the submatrices havae the exact size as the local views.
   * It is not allowed to resize the submatrices.
   * The matrix is initialized and the assembler must only add its contributions to
   * the matrix and not replace values.
   *
   * \param is  An intersection to assembler the local matrix on.
   * \param m   A 2x2 block-matrix to be filled with the local element contributions.
   * \note The local assembler implementation should allow flexible matrix types as argument.
   * \note This function is called for all interior intersections. It is not called
   *       for domain or processor boundary intersections unless this assembler is also
   *       a boundary intersection assembler.
   * \note The user might check `is.neighbor()`
   * \note The outside element is accessible via `is.outside`
   * \note The inside element is equal to the element the assembler was bound to in `bindElement()`.
   *
   * [[pre: m.rows() == 2 && m.cols() == 2]]
   * [[pre: m[0][0].rows() == rowLocalView.size() && m[0][0].cols() == colLocalView.size()]]
   * [[pre: m[0][1].rows() == rowLocalView.size() && m[0][1].cols() == outsideColLocalView.size()]]
   * [[pre: m[1][0].rows() == outsideRowLocalView.size() && m[1][0].cols() == colLocalView.size()]]
   * [[pre: m[1][1].rows() == outsideRowLocalView.size() && m[1][1].cols() == outsideColLocalView.size()]]
   * [[pre: bindElement was called before with current element]]
   */
  template <Concepts::BlockMatrix LocalMatrices>
  void assembleInteriorIntersectionMatrix (const Intersection& is, LocalMatrices& m) noexcept;

  /**
   * Similar as assembleInteriorIntersectionMatrix but the 2x2 block matrix contains local patterns.
   * Each local pattern can be accessed by `p[i][j].add(row,col)`. Additionally, it is possible to
   * set a local pattern to be full, by `p[i][j].addAll()`.
   */
  template <Concepts::BlockLocalMatrixPattern LocalPatterns>
  void assembleInteriorIntersectionMatrixPattern (const Intersection& is, LocalPatterns& p) noexcept;

};

} // namespace Dune::Assembler::Concepts::Archetypes

#endif // DUNE_ASSEMBLER_ARCHETYPES_LOCALINTERIORINTERSECTIONMATRIXASSEMBLER_HH
