// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#ifndef DUNE_ASSEMBLER_ARCHETYPES_LOCALBOUNDARYINTERSECTIONMATRIXASSEMBLER_HH
#define DUNE_ASSEMBLER_ARCHETYPES_LOCALBOUNDARYINTERSECTIONMATRIXASSEMBLER_HH

#include <dune/assembler/concepts.hh>
#include <dune/assembler/localviewpair.hh>

namespace Dune::Assembler::Concepts::Archetypes {

/**
 * Local assembler of a local vector on a boundary intersection.
 *
 * \tparam LocalView   Type of the local view associated to a function space basis.
 *
 * \b Requirements:
 * - `LocalView` is a model of Concepts::LocalView,
 */
template <Concepts::LocalView LocalView>
class LocalBoundaryIntersectionVectorAssembler
{
public:
  using GridView = typename LocalView::GridView;
  using Element = typename GridView::template Codim<0>::Entity;
  using Intersection = typename GridView::Intersection;

public:
  /**
   * This function is called typically before the elements of the grid are traversed.
   * Thus, the local views are not necessarily bound to any element.
   * The lifetime of the passed local view is longer than the lifetime of the local assembler.
   *
   * \param lv   A local views associated to a function space basis
   */
  void bindLocalView (const LocalView& lv) noexcept;

  /**
   * This function is called typically before the elements of the grid are traversed.
   * Thus, the local view is not necessarily bound to any element.
   * The lifetime of the passed local view is longer than the lifetime of the local assembler.
   *
   * \param lv   A local view that will be bound to an
   *             outside element of an intersection.
   */
  void bindOutsideLocalView (const LocalView& lv) noexcept;

  /**
   * When this function is called, the local view is already bound to this element.
   * The lifetime of the passed element is longer than the lifetime of the local assembler.
   *
   * \param e  Grid element the local assembler can be bound to.
   * \note This function is called for all boundary elements. It is not called for inner elements
   *       unless this assembler is also an element or intersection assembler.
   * \note The user might check `e.hasBoundaryIntersections()`
   *
   * [[pre: localView.bound()]]
   */
  void bindElement (const Element& e) noexcept;

  /**
   * The the vector `v` is a 2-component vector `[i, o]` with `i...inside` and `o...outside. Each
   * subvector `v[i]` is assumed to be large enough to be indexable by the corresponding
   * inside/outside local view indices.
   * It is not guaranteed that the subvectors have the exact size as the local views.
   * It is not allowed to resize the subvectors.
   * The vector is already initialized and the assembler must only add its contributions to
   * the vector and not replace values.
   *
   * \param is  An intersection to assembler the local vector on.
   * \param v   A 2 component block-vector to be filled with the local element contributions.
   * \note The local assembler implementation should allow flexible vector types as argument.
   * \note This function is called for all interior intersections. It is not called
   *       for domain or processor boundary intersections unless this assembler is also
   *       a boundary intersection assembler.
   * \note The user might check `is.neighbor()`
   * \note The outside element is accessible via `is.outside`
   * \note The inside element is equal to the element the assembler was bound to in `bindElement()`.
   *
   * [[pre: v.size() == 2]]
   * [[pre: v[0].size() == localView.size()]]
   * [[pre: v[1].size() == outsideLocalView.size()]]
   * [[pre: bindElement was called before with current element]]
   */
  template <Concepts::BlockVector LocalVectors>
  void assembleInteriorIntersectionVector (const Intersection& is, LocalVectors& v) noexcept;
};

} // namespace Dune::Assembler::Concepts::Archetypes

#endif // DUNE_ASSEMBLER_ARCHETYPES_LOCALBOUNDARYINTERSECTIONMATRIXASSEMBLER_HH
