// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#ifndef DUNE_ASSEMBLER_ASSEMBLER_CONCEPTS_HH
#define DUNE_ASSEMBLER_ASSEMBLER_CONCEPTS_HH

#include <dune/assembler/concepts.hh>

namespace Dune::Assembler::Concepts {

// bind to a grid element
template <class LA, class Element>
concept BindableToElement = requires(LA la, const Element& e)
{
  la.bindElement(e);
};

// bind to a single local view
template <class LA, class LocalView>
concept BindableToLocalView =
  Concepts::LocalView<LocalView> &&
  requires(LA la, const LocalView& localView)
{
  la.bindLocalView(localView);
};

// bind to multiple local views
template <class LA, class ... LocalViews>
concept BindableToLocalViews =
  (Concepts::LocalView<LocalViews> &&...) &&
  requires(LA la, const LocalViews&... localViews)
{
  la.bindLocalViews(localViews ...);
};

template <class LA, class RowLocalView, class ColLocalView>
concept LocalElementMatrixAssembler =
  Concepts::BindableToElement<LA, typename RowLocalView::Element> &&
  Concepts::BindableToLocalViews<LA, RowLocalView, ColLocalView> &&
  requires(LA la, typename RowLocalView::Element e, Archetypes::Matrix& localMatrix, Archetypes::LocalMatrixPattern& localPattern)
{
  la.assembleElementMatrix(e,localMatrix);
  la.assembleElementMatrixPattern(e,localPattern);
};


template <class LA, class LocalView>
concept LocalElementVectorAssembler =
  Concepts::BindableToElement<LA, typename LocalView::Element> &&
  Concepts::BindableToLocalView<LA, LocalView> &&
  requires(LA la, typename LocalView::Element e, Archetypes::Vector& localVector)
{
  la.assembleElementVector(e,localVector);
};

template <class LA, class RowLocalView, class ColLocalView>
concept LocalElementSystemAssembler =
  Concepts::BindableToElement<LA, typename RowLocalView::Element> &&
  Concepts::BindableToLocalViews<LA, RowLocalView, ColLocalView> &&
  requires(LA la, typename RowLocalView::Element e, Archetypes::Matrix& localMatrix, Archetypes::Vector& localVector)
{
  la.assembleElement(e,localMatrix,localVector);
};


// assemblers on the boundary intersections

template <class LA, class RowLocalView, class ColLocalView>
concept LocalBoundaryIntersectionMatrixAssembler =
  Concepts::BindableToElement<LA, typename RowLocalView::Element> &&
  Concepts::BindableToLocalViews<LA, RowLocalView, ColLocalView> &&
  requires(LA la, typename RowLocalView::GridView::Intersection is, Archetypes::Matrix& localMatrix, Archetypes::LocalMatrixPattern& localPattern)
{
  la.assembleBoundaryIntersectionMatrix(is,localMatrix);
  la.assembleBoundaryIntersectionMatrixPattern(is,localPattern);
};

template <class LA, class LocalView>
concept LocalBoundaryIntersectionVectorAssembler =
  Concepts::BindableToElement<LA, typename LocalView::Element> &&
  Concepts::BindableToLocalView<LA, LocalView> &&
  requires(LA la, typename LocalView::GridView::Intersection is, Archetypes::Vector& localVector)
{
  la.assembleBoundaryIntersectionVector(is,localVector);
};

template <class LA, class RowLocalView, class ColLocalView>
concept LocalBoundaryIntersectionSystemAssembler =
  Concepts::BindableToElement<LA, typename RowLocalView::Element> &&
  Concepts::BindableToLocalViews<LA, RowLocalView, ColLocalView> &&
  requires(LA la, typename RowLocalView::GridView::Intersection is, Archetypes::Matrix& localMatrix, Archetypes::Vector& localVector)
{
  la.assembleBoundaryIntersection(is,localMatrix,localVector);
};


// assemblers on the inner intersections

// bind to single local view
template <class LA, class LocalView>
concept BindableToOutsideLocalView =
  Concepts::LocalView<LocalView> &&
  requires(LA la, const LocalView& localView)
{
  la.bindOutsideLocalView(localView);
};

// bind to multiple local views
template <class LA, class ... LocalViews>
concept BindableToOutsideLocalViews =
  (Concepts::LocalView<LocalViews> &&...) &&
  requires(LA la, const LocalViews&... localViews)
{
  la.bindOutsideLocalViews(localViews ...);
};

template <class LA, class RowLocalView, class ColLocalView>
concept LocalInteriorIntersectionMatrixAssembler =
  Concepts::BindableToElement<LA, typename RowLocalView::Element> &&
  Concepts::BindableToLocalViews<LA, RowLocalView, ColLocalView> &&
  Concepts::BindableToOutsideLocalViews<LA, RowLocalView, ColLocalView> &&
  requires(LA la, typename RowLocalView::GridView::Intersection is, std::array<std::array<Archetypes::Matrix, 2>, 2> localMatrix, std::array<std::array<Archetypes::LocalMatrixPattern, 2>, 2> localPattern)
{
  la.assembleInteriorIntersectionMatrix(is,localMatrix);
  la.assembleInteriorIntersectionMatrixPattern(is,localPattern);
};

template <class LA, class LocalView>
concept LocalInteriorIntersectionVectorAssembler =
  Concepts::BindableToElement<LA, typename LocalView::Element> &&
  Concepts::BindableToLocalView<LA, LocalView> &&
  Concepts::BindableToOutsideLocalView<LA, LocalView> &&
  requires(LA la, typename LocalView::GridView::Intersection is, std::array<Archetypes::Vector, 2> localVector)
{
  la.assembleInteriorIntersectionVector(is,localVector);
};

template <class LA, class RowLocalView, class ColLocalView>
concept LocalInteriorIntersectionSystemAssembler =
  Concepts::BindableToElement<LA, typename RowLocalView::Element> &&
  Concepts::BindableToLocalViews<LA, RowLocalView, ColLocalView> &&
  Concepts::BindableToOutsideLocalViews<LA, RowLocalView, ColLocalView> &&
  requires(LA la, typename RowLocalView::GridView::Intersection is, std::array<std::array<Archetypes::Matrix, 2>, 2> localMatrix, std::array<Archetypes::Vector, 2> localVector)
{
  la.assembleInteriorIntersection(is,localMatrix,localVector);
};



template <class LA, class RowLocalView, class ColLocalView>
concept LocalMatrixAssembler =
  LocalElementMatrixAssembler<LA, RowLocalView, ColLocalView> ||
  LocalBoundaryIntersectionMatrixAssembler<LA, RowLocalView, ColLocalView> ||
  LocalInteriorIntersectionMatrixAssembler<LA, RowLocalView, ColLocalView>;

template <class LA, class LocalView>
concept LocalVectorAssembler =
  LocalElementVectorAssembler<LA, LocalView> ||
  LocalBoundaryIntersectionVectorAssembler<LA, LocalView> ||
  LocalInteriorIntersectionVectorAssembler<LA, LocalView>;

template <class LA, class RowLocalView, class ColLocalView>
concept LocalSystemAssembler =
  LocalElementSystemAssembler<LA, RowLocalView, ColLocalView> ||
  LocalBoundaryIntersectionSystemAssembler<LA, RowLocalView, ColLocalView> ||
  LocalInteriorIntersectionSystemAssembler<LA, RowLocalView, ColLocalView>;


} // end namespace Dune::Assembler::Concept

#endif // DUNE_ASSEMBLER_ASSEMBLER_CONCEPTS_HH
