// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#ifndef DUNE_ASSEMBLER_BACKENDS_FLATMATRIX_HH
#define DUNE_ASSEMBLER_BACKENDS_FLATMATRIX_HH

#include <vector>

#include <dune/common/std/layout_right.hh>
#include <dune/common/std/mdarray.hh>

namespace Dune::Assembler {

template <class T>
class FlatMatrix
  : public Std::mdarray<T, Std::dextents<std::size_t,2>, Std::layout_right, std::vector<T>>
{
  using Base = Std::mdarray<T, Std::dextents<std::size_t,2>, Std::layout_right, std::vector<T>>;
  using index_type = typename Base::index_type;
  using value_type = typename Base::value_type;
  using extents_type = typename Base::extents_type;
  using mapping_type = typename Base::mapping_type;

  template <class Self>
  struct BracketProxy
  {
    constexpr auto& operator[] (index_type j) const { return (*self_)(i_,j); }
    Self* self_;
    index_type i_;
  };

public:
  using Base::Base;

  constexpr FlatMatrix& operator= (const value_type& value)
  {
    for (index_type i = 0; i < rows(); ++i)
      for (index_type j = 0; j < cols(); ++j)
        (*this)[i][j] = value;
    return *this;
  }

  constexpr BracketProxy<const FlatMatrix> operator[] (index_type i) const { return {this,i}; }
  constexpr BracketProxy<FlatMatrix> operator[] (index_type i) { return {this,i}; }

  constexpr index_type rows () const noexcept { return Base::extent(0); }
  constexpr index_type cols () const noexcept { return Base::extent(1); }

  /// \name Modifiers
  /// @{

  /// \brief Change the shape of the matrix and resize the underlying container with given default value
  void resize (index_type rows, index_type cols, const value_type& value)
  {
    if (rows != this->rows() || cols != this->cols())
    {
      auto container = std::move(*this).extract_container();
      auto m = mapping_type{extents_type{rows,cols}};
      container.resize(m.required_span_size(), value);
      static_cast<Base&>(*this) = Base{m, std::move(container)};
    }
  }

  /// \brief Change the shape of the matrix and resize the underlying container
  void resize (index_type rows, index_type cols)
  {
    if (rows != this->rows() || cols != this->cols())
    {
      auto container = std::move(*this).extract_container();
      auto m = mapping_type{extents_type{rows,cols}};
      container.resize(m.required_span_size());
      static_cast<Base&>(*this) = Base{m, std::move(container)};
    }
  }

  /// @}
};

} // end namespace Dune::Assembler

#endif // DUNE_ASSEMBLER_BACKENDS_FLATMATRIX_HH
