// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#ifndef DUNE_ASSEMBLER_BACKENDS_LOCALPATTERN_HH
#define DUNE_ASSEMBLER_BACKENDS_LOCALPATTERN_HH

#include <cstdint>

#include <dune/assembler/backends/flatmatrix.hh>
#include <dune/assembler/backends/flatset.hh>

namespace Dune::Assembler {

class LocalPatternMatrix
  : public FlatMatrix<char>
{
  using Base = FlatMatrix<char>;

public:
  using Base::Base;
  using size_type = std::size_t;

  void add (size_type row, size_type col)
  {
    (*this)[row][col] = true;
  }

  void addAll ()
  {
    containsAll_ = true;
  }

  [[nodiscard]] bool contains (size_type row, size_type col) const
  {
    return (*this)[row][col];
  }

  [[nodiscard]] bool containsAll () const
  {
    return containsAll_;
  }

  void clear ()
  {
    Base::operator=(0);
    containsAll_ = false;
  }

  template <class RowIndices, class ColIndices, class F>
  void visitEntries (const RowIndices& rowIndices, const ColIndices& colIndices, F&& f) const
  {
    if (containsAll())
      for (auto i : rowIndices)
        for (auto j : colIndices)
          f(i, j);
    else
      for (auto i : rowIndices)
        for (auto j : colIndices)
          if (contains(i,j))
            f(i, j);
  }

private:
  bool containsAll_ = false;
};


class LocalPatternSet
  : public FlatSet<std::pair<std::uint16_t,std::uint16_t>>
{
  using Base = FlatSet<std::pair<std::uint16_t,std::uint16_t>>;

public:
  using value_type = std::pair<std::uint16_t,std::uint16_t>;
  using size_type = std::uint16_t;

public:
  template <class Extents>
  explicit LocalPatternSet (const Extents& extents)
    : LocalPatternSet{extents.extent(0), extents.extent(1)}
  {}

  LocalPatternSet (std::size_t rows, std::size_t cols)
    : rowSize_(rows)
    , colSize_(cols)
  {}

  LocalPatternSet () = default;

  void add (size_type row, size_type col)
  {
    Base::insert(value_type{row,col});
  }

  void addAll ()
  {
    containsAll_ = true;
  }

  void clear ()
  {
    Base::clear();
    containsAll_ = false;
  }

  void resize (std::size_t rowSize, std::size_t colSize)
  {
    rowSize_ = rowSize;
    colSize_ = colSize;
  }

  [[nodiscard]] bool contains (size_type row, size_type col) const
  {
    return Base::contains(value_type{row,col});
  }

  [[nodiscard]] bool containsAll () const
  {
    return containsAll_;
  }

  template <class RowIndices, class ColIndices, class F>
  void visitEntries (const RowIndices& rowIndices, const ColIndices& colIndices, F&& f) const
  {
    if (containsAll())
      for (auto i : rowIndices)
        for (auto j : colIndices)
          f(i, j);
    else
      for (auto&& [i,j] : *this)
        f(i, j);
  }

private:
  std::size_t rowSize_ = 0;
  std::size_t colSize_ = 0;

  bool containsAll_ = false;
};

} // end namespace Dune::Assembler

#endif // DUNE_ASSEMBLER_BACKENDS_LOCALPATTERN_HH
