// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#ifndef DUNE_ASSEMBLER_DENSELOCALASSEMBLERBASE_HH
#define DUNE_ASSEMBLER_DENSELOCALASSEMBLERBASE_HH

#include <dune/assembler/concepts.hh>

namespace Dune::Assembler {


/**
 * \brief Base class for local assemblers with dense local pattern
 *
 * This is not a mandatory base class but adds default implementations
 * of the various `assemble*Pattern(..., localPattern)` methods
 * that fill the local pattern using `localPattern.addall()`.
 */
template <Concepts::LocalView RowLocalView, Concepts::LocalView ColLocalView = RowLocalView>
class DenseLocalAssemblerBase
{
  using GridView = typename RowLocalView::GridView;
  using Element = typename GridView::template Codim<0>::Entity;
  using Intersection = typename GridView::Intersection;

public:
  /**
   * Default implementation for a dense pattern.
   * This default implementation sets a dense pattern by
   * calling `localPattern.addAll()`.
   */
  template <Concepts::LocalMatrixPattern LocalPattern>
  void assembleElementMatrixPattern (const Element& e, LocalPattern& p) noexcept
  {
    p.addAll();
  }

  /**
   * Default implementation for a dense pattern.
   * This default implementation sets a dense pattern by
   * calling `localPattern.addAll()`.
   */
  template <Concepts::LocalMatrixPattern LocalPattern>
  void assembleBoundaryIntersectionMatrixPattern (const Intersection& is, LocalPattern& p) noexcept
  {
    p.addAll();
  }

  /**
   * Default implementation for a dense pattern.
   * This default implementation sets a dense pattern by
   * calling `localPattern.addAll()`.
   */
  template <Concepts::BlockLocalMatrixPattern LocalPatterns>
  void assembleInteriorIntersectionMatrixPattern (const Intersection& is, LocalPatterns& p) noexcept
  {
    p[0][0].addAll();
    p[0][1].addAll();
    p[1][0].addAll();
    p[1][1].addAll();
  }

};


} // end namespace Dune::Assembler

#endif // DUNE_ASSEMBLER_DENSELOCALASSEMBLERBASE_HH
