// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#ifndef DUNE_ASSEMBLER_LOCALVIEWPAIR_HH
#define DUNE_ASSEMBLER_LOCALVIEWPAIR_HH

#include <type_traits>

#include <dune/assembler/concepts.hh>
#include <dune/common/std/extents.hh>

namespace Dune::Assembler {

/**
 * \brief A pair of local views constructed from a pair of global bases.
 *
 * The pair makes sure that in case of identical bases only one object is filled
 * and initialized. Of the bases are identical, the second local view is just a
 * reference to the first local view. But if the bases are different, two separate
 * local views will be handled.
 */
template <Concepts::LocalView RowLocalView, Concepts::LocalView ColLocalView>
class LocalViewPair
{
  using RowBasis = typename RowLocalView::GlobalBasis;
  using ColBasis = typename ColLocalView::GlobalBasis;

  using size_type = std::common_type_t<typename RowLocalView::size_type, typename ColLocalView::size_type>;

  // create a reference to the first local view of the bases and the local views
  // are identical, otherwise return a reference to the second local view.
  template <class LV1, class LV2>
  static auto& refColLocalView (LV1& first, LV2& second, bool same) noexcept
  {
    if constexpr (std::is_same_v<LV1,LV2>)
      return same ? first : second;
    else
      return second;
  }

public:
  /// \brief Construct local view for a given global finite element basis
  LocalViewPair (RowBasis const& rowBasis, ColBasis const& colBasis)
    : rowLocalView_(rowBasis.localView())
    , colLocalViewStorage_(colBasis.localView())
    , colLocalView_(refColLocalView(rowLocalView_, colLocalViewStorage_,
                                    (void*)(&rowBasis) == (void*)(&colBasis)))
  {}

  /**
   * \brief Copy-construct a LocalView pair.
   * This needs to reset the reference for the second local view.
   **/
  LocalViewPair (const LocalViewPair& other)
    : rowLocalView_(other.rowLocalView_)
    , colLocalViewStorage_(other.colLocalViewStorage_)
    , colLocalView_(refColLocalView(rowLocalView_, colLocalViewStorage_, other.same()))
  {}

  /**
   * \brief Move-construct a LocalView pair.
   * This needs to reset the reference for the second local view.
   **/
  LocalViewPair (LocalViewPair&& other)
    : rowLocalView_(std::move(other.rowLocalView_))
    , colLocalViewStorage_(std::move(other.colLocalViewStorage_))
    , colLocalView_(refColLocalView(rowLocalView_, colLocalViewStorage_, other.same()))
  {}

  /// \brief Bind the views to a grid element
  template <class Element>
  void bind (const Element& element)
  {
    rowLocalView_.bind(element);
    if (not same())
      colLocalView_.bind(element);
  }

  /// \brief Unbind from the current element
  void unbind ()
  {
    rowLocalView_.unbind();
    if (not same())
      colLocalView_.unbind();
  }

  /// \brief Return a reference to the first (I == 0) or second (I == 1) local view.
  template <std::size_t I>
  friend auto const& get (const LocalViewPair& pair) noexcept
  {
    if constexpr (I == 0)
      return pair.rowLocalView_;
    else
      return pair.colLocalView_;
  }

  /**
   * \brief Return whether both local views refer to the same object.
   * \note This is equivalent to `&get<0>(*this) == &get<1>(*this)`.
   */
  [[nodiscard]] bool same () const
  {
    return ((void*)(&rowLocalView_) == (void*)(&colLocalView_));
  }

protected:
  RowLocalView rowLocalView_;
  ColLocalView colLocalViewStorage_;
  ColLocalView& colLocalView_;
};

template <Concepts::GlobalBasis RowBasis, Concepts::GlobalBasis ColBasis>
LocalViewPair(const RowBasis&, const ColBasis&)
-> LocalViewPair<typename RowBasis::LocalView, typename ColBasis::LocalView>;

} // end namespace Dune::Assembler

#endif // DUNE_ASSEMBLER_LOCALVIEWPAIR_HH
