// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#include <config.h>

#include <petscsys.h>
#include <petscksp.h>


#include <dune/assembler/defaultglobalassembler.hh>
#include <dune/assembler/backends/petscmatrixbackend.hh>
#include <dune/assembler/backends/petscvectorbackend.hh>

#include <dune/common/parallel/mpihelper.hh>
#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/grid/yaspgrid.hh>

#include "massmatrixlocalassembler.hh"

int main (int argc, char** argv)
{
  Dune::MPIHelper::instance(argc, argv);
  PetscInitialize(&argc, &argv, PETSC_NULL, PETSC_NULL);

  auto grid = Dune::YaspGrid<2>({1.0,1.0}, {20,20});
  auto gridView = grid.leafGridView();

  using namespace Dune::Functions::BasisFactory;
  auto basis = makeBasis(gridView, lagrange<1>());
  using Basis = decltype(basis);

  auto assembler = Dune::Assembler::Assembler{basis};

  using LocalView = typename Basis::LocalView;
  auto localAssembler = Dune::Assembler::MassMatrixLocalAssembler<LocalView,LocalView>{};

  Mat massMatrix{};
  MatCreate(PETSC_COMM_SELF, &massMatrix);
  auto matrixBackend = Dune::Assembler::PETScMatrixBackend{massMatrix};
  auto patternBuilder = matrixBackend.patternBuilder();
  patternBuilder.resize(basis,basis);

  // assemble the sparsity pattern
  assembler.assembleMatrixPattern(localAssembler, patternBuilder);
  patternBuilder.setupMatrix();

  // set all entries to zero
  assembler.assembleMatrixEntries(localAssembler, matrixBackend);
  matrixBackend.finish();

  Vec b{};
  VecCreate(PETSC_COMM_SELF, &b);
  auto bBackend = Dune::Assembler::PETScVectorBackend{b};
  bBackend.resize(basis);
  bBackend = 1.0;

  Vec u{};
  VecCreate(PETSC_COMM_SELF, &u);
  auto uBackend = Dune::Assembler::PETScVectorBackend{u};
  uBackend.resize(basis);
  uBackend.setZero();

  // solve the linear system
  KSP ksp{};
  KSPCreate(PETSC_COMM_SELF, &ksp);
  KSPSetOperators(ksp, massMatrix, massMatrix);
  KSPSetType(ksp, KSPCG);
  KSPSetFromOptions(ksp);

  PC pc{};
  KSPGetPC(ksp, &pc);
  PCSetType(pc, PCJACOBI);
  PCSetUp(pc);

  // solve the linear system A*u = b
  KSPSolve(ksp, b, u);
  KSPConvergedReasonView(ksp, PETSC_VIEWER_STDOUT_WORLD);

  KSPDestroy(&ksp);
}
