// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#pragma once

#include <cstddef>
#include <optional>
#include <type_traits>
#include <vector>

#include <dune/geometry/quadraturerules.hh>
#include <dune/assembler/denselocalassemblerbase.hh>

namespace Dune::Assembler {

template <class TestLocalView, class TrialLocalView>
class MassMatrixLocalAssembler
  : public Dune::Assembler::DenseLocalAssemblerBase<TestLocalView, TrialLocalView>
{
  using GridView = typename TestLocalView::GridView;
  using Element = typename TestLocalView::Element;
  using Geometry = typename Element::Geometry;

  using QuadRules = Dune::QuadratureRules<typename GridView::ctype, GridView::dimension>;
  using QuadRule = Dune::QuadratureRule<typename GridView::ctype, GridView::dimension>;

public:
  /**
   * \param test The local-view of the test-function space (not yet bound to any element)
   */
  void bindLocalViews (const TestLocalView& testLocalView, const TrialLocalView& trialLocalView)
  {
    testLocalView_ = &testLocalView;
    trialLocalView_ = &trialLocalView;
  }


  /**
   * \param element  The grid-element this local-assembler needs to be bound to
   */
  void bindElement (const Element& element)
  {
    quadRule_ = &QuadRules::rule(element.type(), 4);
    geometry_.emplace(element.geometry());
  }

  /**
   * \brief Fill the element matrix by assembling on each leaf-basis node a mass matrix.
   * \note No optimization regarding equal spaces is implemented here for simplicity.
   */
  template <class LocalMatrix>
  void assembleElementMatrix (const Element& element, LocalMatrix& localMatrix) const
  {
    assert(testLocalView_->tree().isLeaf && trialLocalView_->tree().isLeaf);
    const auto& testNode = testLocalView_->tree();
    const auto& trialNode = trialLocalView_->tree();

    const auto& testFE = testNode.finiteElement();
    const auto& testLB = testFE.localBasis();
    using TestLocalBasisType = std::decay_t<decltype(testLB)>;
    std::vector<typename TestLocalBasisType::Traits::RangeType> testShapeValues;

    const auto& trialFE = trialNode.finiteElement();
    const auto& trialLB = trialFE.localBasis();
    using TrialLocalBasisType = std::decay_t<decltype(trialLB)>;
    std::vector<typename TrialLocalBasisType::Traits::RangeType> trialShapeValues;

    assert(!!quadRule_);
    for (auto&& [pos, weight] : *quadRule_)
    {
      auto dx = geometry_->integrationElement(pos) * weight;

      testLB.evaluateFunction(pos, testShapeValues);
      trialLB.evaluateFunction(pos, trialShapeValues);

      for (std::size_t i = 0; i < testShapeValues.size(); ++i)
      {
        const auto local_i = testNode.localIndex(i);
        for (std::size_t j = 0; j < trialShapeValues.size(); ++j)
        {
          const auto local_j = trialNode.localIndex(j);
          localMatrix[local_i][local_j] += testShapeValues[i] * trialShapeValues[j] * dx;
        }
      }
    }
  }

private:
  const TestLocalView* testLocalView_ = nullptr;
  const TrialLocalView* trialLocalView_ = nullptr;

  const QuadRule* quadRule_ = nullptr;
  std::optional<Geometry> geometry_ = std::nullopt;
};

} // end namespace Dune::Assembler
