// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#include <config.h>

#include <vector>
#include <cmath>

#include <dune/common/bitsetvector.hh>
#include <dune/common/rangeutilities.hh>
#include <dune/common/timer.hh>

#include <dune/geometry/quadraturerules.hh>

#include <dune/grid/yaspgrid.hh>
#include <dune/grid/uggrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>

#include <dune/grid/io/file/gmshreader.hh>

#include <dune/functions/functionspacebases/interpolate.hh>
#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/functions/functionspacebases/boundarydofs.hh>
#include <dune/functions/gridfunctions/discreteglobalbasisfunction.hh>
#include <dune/functions/gridfunctions/gridviewfunction.hh>

#include <dune/assembler/backends/eigenbackend.hh>
#include <dune/assembler/defaultglobalassembler.hh>
#include <dune/assembler/denselocalassemblerbase.hh>

#include <Eigen/Sparse>
#include <Eigen/SparseCholesky>

using namespace Dune;


template <class LocalView, class RHSFunc>
class LocalPoissonAssembler
  : public Dune::Assembler::DenseLocalAssemblerBase<LocalView>
{
  using GridView = typename LocalView::GridView;
  using Element = typename LocalView::Element;
  using LocalRHSFunc = typename RHSFunc::LocalFunction;

  static const int dim = Element::dimension;

public:

  LocalPoissonAssembler (const RHSFunc& rhs)
    : rhs_(rhs),
    localRhs_(localFunction(rhs))
  {}

  template <class Basis>
  LocalPoissonAssembler (const Basis&, const RHSFunc& rhs)
    : rhs_(rhs),
    localRhs_(localFunction(rhs))
  {}

  void bindLocalViews (const LocalView& testLocalView, const LocalView& trialLocalView)
  {
    testLocalView_ = &testLocalView;
    trialLocalView_ = &trialLocalView;
  }

  void bindElement (const Element& element)
  {}

  template <class LocalMatrix>
  void assembleElementMatrix (const Element& element, LocalMatrix& localMatrix)
  {
    auto geometry = element.geometry();

    const auto& node = testLocalView_->tree();
    const auto& finiteElement = node.finiteElement();
    referenceJacobians_.resize(finiteElement.size());
    jacobians_.resize(finiteElement.size());

    int order = 2*finiteElement.localBasis().order();
    const auto& quad = Dune::QuadratureRules<double, dim>::rule(element.type(), order);

    for (const auto& quadPoint : quad)
    {
      const auto integrationWeigth = geometry.integrationElement(quadPoint.position()) * quadPoint.weight();

      const auto jacobianInverse = geometry.jacobianInverse(quadPoint.position());
      finiteElement.localBasis().evaluateJacobian(quadPoint.position(), referenceJacobians_);
      for (auto i : Dune::range(finiteElement.size()))
        jacobians_[i] = referenceJacobians_[i] * jacobianInverse;

      for (auto i : Dune::range(finiteElement.size()))
        for (auto j : Dune::range(finiteElement.size()))
          localMatrix[node.localIndex(i)][node.localIndex(j)] += (jacobians_[i] * transpose(jacobians_[j])) * integrationWeigth;
    }
  }

  void bindLocalView (const LocalView& testLocalView)
  {
    testLocalView_ = &testLocalView;
  }

  template <class LocalVector>
  void assembleElementVector (const Element& element, LocalVector& localVector)
  {
    auto geometry = element.geometry();

    localRhs_.bind(element);

    const auto& node = testLocalView_->tree();
    const auto& finiteElement = node.finiteElement();
    values_.resize(finiteElement.size());

    int order = 2*finiteElement.localBasis().order();
    const auto& quad = Dune::QuadratureRules<double, dim>::rule(element.type(), order);

    for (const auto& quadPoint : quad)
    {
      const auto integrationWeigth = geometry.integrationElement(quadPoint.position()) * quadPoint.weight();

      finiteElement.localBasis().evaluateFunction(quadPoint.position(), values_);

      auto rhsValue = localRhs_(quadPoint.position());

      for (auto i : Dune::range(finiteElement.size()))
        localVector[node.localIndex(i)] += values_[i] * rhsValue * integrationWeigth;
    }
  }

private:
  RHSFunc rhs_;
  LocalRHSFunc localRhs_;
  const LocalView* testLocalView_ = nullptr;
  const LocalView* trialLocalView_ = nullptr;
  std::vector<Dune::FieldMatrix<double,1,dim>> referenceJacobians_;
  std::vector<Dune::FieldMatrix<double,1,dim>> jacobians_;
  std::vector<Dune::FieldVector<double,1>> values_;
};

template <class Basis, class RHSFunc>
LocalPoissonAssembler(const Basis&, const RHSFunc& rhs) -> LocalPoissonAssembler<typename Basis::LocalView, RHSFunc>;



// This method marks all Lagrange nodes on the boundary of the grid.
// In our problem we want to use them as the Dirichlet nodes.
// The result can be found in the 'dirichletNodes' variable.  There, a bit
// is set precisely when the corresponding Lagrange node is on the grid boundary.
//
// Since interpolating into a vector<bool> is currently not supported,
// we use a vector<char> which, in contrast to vector<bool>
// is a real container.
template <class FEBasis>
void boundaryTreatment (const FEBasis& feBasis, std::vector<char>& dirichletNodes )
{
  dirichletNodes.clear();
  dirichletNodes.resize(feBasis.size(), false);

  Functions::forEachBoundaryDOF(feBasis, [&] (auto&& index) {
    dirichletNodes[index] = true;
  });
}

auto createMixedGrid ()
{
  using Grid = Dune::UGGrid<2>;
  Dune::GridFactory<Grid> factory;
  for (unsigned int k : Dune::range(9))
    factory.insertVertex({0.5*(k%3), 0.5*(k/3)});
  factory.insertElement(Dune::GeometryTypes::cube(2), {0, 1, 3, 4});
  factory.insertElement(Dune::GeometryTypes::cube(2), {1, 2, 4, 5});
  factory.insertElement(Dune::GeometryTypes::simplex(2), {3, 4, 6});
  factory.insertElement(Dune::GeometryTypes::simplex(2), {4, 7, 6});
  factory.insertElement(Dune::GeometryTypes::simplex(2), {4, 5, 7});
  factory.insertElement(Dune::GeometryTypes::simplex(2), {5, 8, 7});
  return factory.createGrid();
}

int main (int argc, char *argv[]) try
{
  // Set up MPI, if available
  MPIHelper::instance(argc, argv);

  std::size_t refinements = 4;
  if (argc>1)
    refinements = std::stoul(std::string(argv[1]));

  ///////////////////////////////////
  //   Generate the grid
  ///////////////////////////////////

  auto grid = createMixedGrid();

  grid->globalRefine(refinements);

  auto gridView = grid->leafGridView();
  using GridView = decltype(gridView);

  /////////////////////////////////////////////////////////
  //   Choose a finite element space
  /////////////////////////////////////////////////////////

  using FEBasis = Functions::LagrangeBasis<GridView,2>;
  auto feBasis = FEBasis(gridView);

  /////////////////////////////////////////////////////////
  //   Stiffness matrix and right hand side vector
  /////////////////////////////////////////////////////////

  using Vector = Eigen::Matrix<double, Eigen::Dynamic, 1>;
  using BitVector = std::vector<char>;
  using Matrix = Eigen::SparseMatrix<double, Eigen::RowMajor>;

  auto rhs = Vector();
  auto stiffnessMatrix = Matrix();

  /////////////////////////////////////////////////////////
  //  Assemble the system
  /////////////////////////////////////////////////////////

  std::cout << "Number of DOFs is " << feBasis.dimension() << std::endl;

  auto rightHandSide = [] (const auto& x) {
    return 10;
  };

  Dune::Timer timer;


  auto matrixBackend = Dune::Assembler::EigenMatrixBackend(stiffnessMatrix);
  auto patternBuilder = matrixBackend.patternBuilder();
  auto rhsBackend = Dune::Assembler::EigenVectorBackend(rhs);

  auto assembler = Dune::Assembler::Assembler (feBasis);

  auto localAssembler = LocalPoissonAssembler(feBasis, Dune::Functions::makeGridViewFunction(rightHandSide, feBasis.gridView()));

  patternBuilder.resize(feBasis, feBasis);

  assembler.assembleMatrixPattern(localAssembler, patternBuilder);

  patternBuilder.setupMatrix();

  rhsBackend.resize(feBasis);
  rhsBackend.setZero();

  assembler.assembleMatrixEntries(localAssembler, matrixBackend);
  assembler.assembleVectorEntries(localAssembler, rhsBackend);

  std::cout << "Assembling the problem took " << timer.elapsed() << "s" << std::endl;

  /////////////////////////////////////////////////
  //   Choose an initial iterate
  /////////////////////////////////////////////////
  auto x = Vector();

  Dune::Assembler::EigenVectorBackend(x).resize(feBasis);
  Dune::Assembler::EigenVectorBackend(x).setZero();

  // Determine Dirichlet dofs
  auto dirichletNodes = BitVector();
  boundaryTreatment(feBasis, dirichletNodes);



  // Don't trust on non-standard M_PI.
  auto pi = std::acos(-1.0);
  auto dirichletValueFunction = [pi](const auto& x){
    return std::sin(2*pi*x[0]);
  };

  // Interpolate dirichlet values at the boundary nodes
  interpolate(feBasis, x, dirichletValueFunction, dirichletNodes);

  //////////////////////////////////////////////////////
  //   Incorporate Dirichlet values in a symmetric way
  //////////////////////////////////////////////////////

  // Compute residual for non-homogeneous Dirichlet values
  // stored in x. Since x is zero for non-Dirichlet DOFs,
  // we can simply multiply by the matrix.
  rhs -= stiffnessMatrix*x;

  // Change Dirichlet matrix rows and columns to the identity.
  for (size_t i=0; i<stiffnessMatrix.rows(); i++)
  {
    if (dirichletNodes[i])
    {
      rhs[i] = x[i];
      // loop over nonzero matrix entries in current row
      for ( auto it = Matrix::InnerIterator(stiffnessMatrix, i); it; ++it)
        it.valueRef() = (i==it.col()) ? 1.0 : 0.0;
    }
    else
    {
      // loop over nonzero matrix entries in current row
      for ( auto it = Matrix::InnerIterator(stiffnessMatrix, i); it; ++it)
        if (dirichletNodes[it.col()])
          it.valueRef() = 0.0;
    }
  }

  ////////////////////////////
  //   Compute solution
  ////////////////////////////

  auto llt = Eigen::SimplicialLLT<Matrix>();

  llt.compute(stiffnessMatrix);
  x = llt.solve(rhs);

  ////////////////////////////////////////////////////////////////////////////
  //  Make a discrete function from the FE basis and the coefficient vector
  ////////////////////////////////////////////////////////////////////////////

  auto xFunction = Functions::makeDiscreteGlobalBasisFunction<double>(feBasis, x);

  //////////////////////////////////////////////////////////////////////////////////////////////
  //  Write result to VTK file
  //  We need to subsample, because VTK cannot natively display real second-order functions
  //////////////////////////////////////////////////////////////////////////////////////////////
  SubsamplingVTKWriter<GridView> vtkWriter(gridView, Dune::refinementLevels(2));
  vtkWriter.addVertexData(xFunction, VTK::FieldInfo("x", VTK::FieldInfo::Type::scalar, 1));
  vtkWriter.write("poisson-pq2-eigen");

}
// Error handling
catch (Exception& e) {
  std::cout << e.what() << std::endl;
}
