// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#include <config.h>

#include <numbers>
#include <vector>
#include <cmath>

#include <dune/common/bitsetvector.hh>
#include <dune/common/rangeutilities.hh>

#include <dune/geometry/quadraturerules.hh>

#include <dune/grid/yaspgrid.hh>
#include <dune/grid/uggrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>

#include <dune/grid/io/file/gmshreader.hh>

#include <dune/istl/matrix.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/matrixindexset.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/solvers.hh>

#include <dune/functions/functionspacebases/interpolate.hh>
#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/functions/functionspacebases/boundarydofs.hh>
#include <dune/functions/gridfunctions/discreteglobalbasisfunction.hh>
#include <dune/functions/gridfunctions/gridviewfunction.hh>

#include <dune/assembler/defaultglobalassembler.hh>
#include <dune/assembler/denselocalassemblerbase.hh>
#include <dune/assembler/backends/petscmatrixbackend.hh>
#include <dune/assembler/backends/petscvectorbackend.hh>

#include <petscmat.h>
#include <petscsys.h>
#include <petscksp.h>

using namespace Dune;


template <class LocalView, class RHSFunc>
class LocalPoissonAssembler
  : public Dune::Assembler::DenseLocalAssemblerBase<LocalView>
{
  using GridView = typename LocalView::GridView;
  using Element = typename LocalView::Element;
  using LocalRHSFunc = typename RHSFunc::LocalFunction;

  static const int dim = Element::dimension;

public:

  LocalPoissonAssembler (const RHSFunc& rhs)
    : rhs_(rhs),
    localRhs_(localFunction(rhs))
  {}

  template <class Basis>
  LocalPoissonAssembler (const Basis&, const RHSFunc& rhs)
    : rhs_(rhs),
    localRhs_(localFunction(rhs))
  {}

  void bindLocalViews (const LocalView& testLocalView, const LocalView& trialLocalView)
  {
    testLocalView_ = &testLocalView;
    trialLocalView_ = &trialLocalView;
  }

  void bindElement (const Element& element)
  {}

  template <class LocalMatrix>
  void assembleElementMatrix (const Element& element, LocalMatrix& localMatrix)
  {
    auto geometry = element.geometry();

    const auto& node = testLocalView_->tree();
    const auto& finiteElement = node.finiteElement();
    referenceJacobians_.resize(finiteElement.size());
    jacobians_.resize(finiteElement.size());

    int order = 2*finiteElement.localBasis().order();
    const auto& quad = Dune::QuadratureRules<double, dim>::rule(element.type(), order);

    for (const auto& quadPoint : quad)
    {
      const auto integrationWeigth = geometry.integrationElement(quadPoint.position()) * quadPoint.weight();

      const auto jacobianInverse = geometry.jacobianInverse(quadPoint.position());
      finiteElement.localBasis().evaluateJacobian(quadPoint.position(), referenceJacobians_);
      for (auto i : Dune::range(finiteElement.size()))
        jacobians_[i] = referenceJacobians_[i] * jacobianInverse;

      for (auto i : Dune::range(finiteElement.size()))
        for (auto j : Dune::range(finiteElement.size()))
          localMatrix[node.localIndex(i)][node.localIndex(j)] += (jacobians_[i] * transpose(jacobians_[j])) * integrationWeigth;
    }
  }

  void bindLocalView (const LocalView& testLocalView)
  {
    testLocalView_ = &testLocalView;
  }

  template <class LocalVector>
  void assembleElementVector (const Element& element, LocalVector& localVector)
  {
    auto geometry = element.geometry();

    localRhs_.bind(element);

    const auto& node = testLocalView_->tree();
    const auto& finiteElement = node.finiteElement();
    values_.resize(finiteElement.size());

    int order = 2*finiteElement.localBasis().order();
    const auto& quad = Dune::QuadratureRules<double, dim>::rule(element.type(), order);

    for (const auto& quadPoint : quad)
    {
      const auto integrationWeigth = geometry.integrationElement(quadPoint.position()) * quadPoint.weight();

      finiteElement.localBasis().evaluateFunction(quadPoint.position(), values_);

      auto rhsValue = localRhs_(quadPoint.position());

      for (auto i : Dune::range(finiteElement.size()))
        localVector[node.localIndex(i)] += values_[i] * rhsValue * integrationWeigth;
    }
  }

private:
  RHSFunc rhs_;
  LocalRHSFunc localRhs_;
  const LocalView* testLocalView_ = nullptr;
  const LocalView* trialLocalView_ = nullptr;
  std::vector<Dune::FieldMatrix<double,1,dim>> referenceJacobians_;
  std::vector<Dune::FieldMatrix<double,1,dim>> jacobians_;
  std::vector<Dune::FieldVector<double,1>> values_;
};

template <class Basis, class RHSFunc>
LocalPoissonAssembler(const Basis&, const RHSFunc& rhs) -> LocalPoissonAssembler<typename Basis::LocalView, RHSFunc>;



// This method marks all Lagrange nodes on the boundary of the grid.
// In our problem we want to use them as the Dirichlet nodes.
// The result can be found in the 'dirichletNodes' variable.  There, a bit
// is set precisely when the corresponding Lagrange node is on the grid boundary.
//
// Since interpolating into a vector<bool> is currently not supported,
// we use a vector<char> which, in contrast to vector<bool>
// is a real container.
template <class FEBasis>
void boundaryTreatment (const FEBasis& feBasis, std::vector<char>& dirichletNodes )
{
  dirichletNodes.clear();
  dirichletNodes.resize(feBasis.size(), false);

  Functions::forEachBoundaryDOF(feBasis, [&] (auto&& index) {
    dirichletNodes[index] = true;
  });
}

auto createMixedGrid ()
{
  using Grid = Dune::UGGrid<2>;
  Dune::GridFactory<Grid> factory;
  for (unsigned int k : Dune::range(9))
    factory.insertVertex({0.5*(k%3), 0.5*(k/3)});
  factory.insertElement(Dune::GeometryTypes::cube(2), {0, 1, 3, 4});
  factory.insertElement(Dune::GeometryTypes::cube(2), {1, 2, 4, 5});
  factory.insertElement(Dune::GeometryTypes::simplex(2), {3, 4, 6});
  factory.insertElement(Dune::GeometryTypes::simplex(2), {4, 7, 6});
  factory.insertElement(Dune::GeometryTypes::simplex(2), {4, 5, 7});
  factory.insertElement(Dune::GeometryTypes::simplex(2), {5, 8, 7});
  return factory.createGrid();
}

int main (int argc, char *argv[]) try
{
  // Set up MPI, if available
  MPIHelper::instance(argc, argv);
  PetscInitialize(&argc, &argv, PETSC_NULL, PETSC_NULL);

  std::size_t refinements = 4;
  if (argc>1)
    refinements = std::stoul(std::string(argv[1]));

  ///////////////////////////////////
  //   Generate the grid
  ///////////////////////////////////

  auto grid = createMixedGrid();

  grid->globalRefine(refinements);

  auto gridView = grid->leafGridView();
  using GridView = decltype(gridView);

  /////////////////////////////////////////////////////////
  //   Choose a finite element space
  /////////////////////////////////////////////////////////

  using FEBasis = Functions::LagrangeBasis<GridView,2>;
  auto feBasis = FEBasis{gridView};

  /////////////////////////////////////////////////////////
  //   Stiffness matrix and right hand side vector
  /////////////////////////////////////////////////////////

  Vec rhs{};
  VecCreate(PETSC_COMM_SELF, &rhs);
  auto rhsBackend = Dune::Assembler::PETScVectorBackend{rhs};

  Mat stiffnessMatrix{};
  MatCreate(PETSC_COMM_SELF, &stiffnessMatrix);
  auto matrixBackend = Dune::Assembler::PETScMatrixBackend{stiffnessMatrix};

  /////////////////////////////////////////////////////////
  //  Assemble the system
  /////////////////////////////////////////////////////////

  std::cout << "Number of DOFs is " << feBasis.dimension() << std::endl;

  auto rightHandSide = [] (const auto& x) {
    return 10;
  };

  Dune::Timer timer;

  auto assembler = Dune::Assembler::Assembler<FEBasis>(feBasis);
  auto localAssembler
    = LocalPoissonAssembler(feBasis, Dune::Functions::makeGridViewFunction(rightHandSide, feBasis.gridView()));

  auto patternBuilder = matrixBackend.patternBuilder();
  patternBuilder.resize(feBasis, feBasis);
  assembler.assembleMatrixPattern(localAssembler, patternBuilder);
  patternBuilder.setupMatrix();

  assembler.assembleMatrixEntries(localAssembler, matrixBackend);
  matrixBackend.finish();
  MatSetOption(stiffnessMatrix, MAT_SPD, PETSC_TRUE);

  rhsBackend.resize(feBasis);
  rhsBackend.setZero();
  assembler.assembleVectorEntries(localAssembler, rhsBackend);
  rhsBackend.finish();

  std::cout << "Assembling the problem took " << timer.elapsed() << "s" << std::endl;
  std::cout << "Number of nonzeros: " << matrixBackend.nnz() << std::endl;
  timer.reset();

  /////////////////////////////////////////////////
  //   Choose an initial iterate
  /////////////////////////////////////////////////

  // Determine Dirichlet dofs
  std::vector<char> dirichletNodes;
  boundaryTreatment(feBasis, dirichletNodes);

  auto dirichletValueFunction = [](const auto& x) {
    return std::sin(2*std::numbers::pi*x[0]);
  };

  Vec x{};
  VecCreate(PETSC_COMM_SELF, &x);
  auto xBackend = Dune::Assembler::PETScVectorBackend{x};
  xBackend.resize(feBasis);
  xBackend.setZero();
  interpolate(feBasis, xBackend, dirichletValueFunction, dirichletNodes);

  // transform the dirichlet nodes into a list of indice
  std::vector<PetscInt> dirichletIndices;
  dirichletIndices.reserve(std::accumulate(dirichletNodes.begin(), dirichletNodes.end(), 0));
  for (std::size_t i = 0; i < dirichletNodes.size(); ++i)
    if (dirichletNodes[i])
      dirichletIndices.push_back(i);

  //////////////////////////////////////////////////////
  //   Incorporate Dirichlet values in a symmetric way
  //////////////////////////////////////////////////////

  // Change Dirichlet matrix rows and columns to the identity.
  MatZeroRowsColumns(stiffnessMatrix, dirichletIndices.size(), dirichletIndices.data(), 1.0, x, rhs);
  std::cout << "Assembling the boundary conditions took " << timer.elapsed() << "s" << std::endl;


  ////////////////////////////
  //   Compute solution
  ////////////////////////////

  KSP ksp{};
  KSPCreate(PETSC_COMM_SELF, &ksp);
  KSPSetOperators(ksp, stiffnessMatrix, stiffnessMatrix);
  KSPSetType(ksp, KSPCG);
  KSPSetInitialGuessNonzero(ksp, PETSC_TRUE);
  KSPSetFromOptions(ksp);

  PC pc{};
  KSPGetPC(ksp, &pc);
  PCSetType(pc, PCICC);
  PCSetUp(pc);

  // solve the linear system A*u = b
  KSPSolve(ksp, rhs, x);
  KSPConvergedReasonView(ksp, PETSC_VIEWER_STDOUT_WORLD);

  KSPDestroy(&ksp);

  ////////////////////////////////////////////////////////////////////////////
  //  Make a discrete function from the FE basis and the coefficient vector
  ////////////////////////////////////////////////////////////////////////////

  auto xFunction = Functions::makeDiscreteGlobalBasisFunction<double>(feBasis, xBackend);

  //////////////////////////////////////////////////////////////////////////////////////////////
  //  Write result to VTK file
  //  We need to subsample, because VTK cannot natively display real second-order functions
  //////////////////////////////////////////////////////////////////////////////////////////////

  SubsamplingVTKWriter<GridView> vtkWriter(gridView, Dune::refinementLevels(2));
  vtkWriter.addVertexData(xFunction, VTK::FieldInfo("x", VTK::FieldInfo::Type::scalar, 1));
  vtkWriter.write("poisson-pq2-petsc");

}
// Error handling
catch (Exception& e) {
  std::cout << e.what() << std::endl;
}
