// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

// SPDX-FileCopyrightText: Copyright © DUNE Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#include <config.h>

#include <array>
#include <vector>

#include <dune/common/bitsetvector.hh>
#include <dune/common/indices.hh>
#include <dune/common/transpose.hh>

#include <dune/geometry/quadraturerules.hh>

#include <dune/grid/yaspgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>

#include <dune/istl/matrix.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/matrixindexset.hh>
#include <dune/istl/solvers.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/multitypeblockmatrix.hh>
#include <dune/istl/multitypeblockvector.hh>
#include <dune/istl/foreach.hh>

#include <dune/functions/functionspacebases/interpolate.hh>
#include <dune/functions/functionspacebases/taylorhoodbasis.hh>
#include <dune/functions/backends/istlvectorbackend.hh>
#include <dune/functions/functionspacebases/powerbasis.hh>
#include <dune/functions/functionspacebases/compositebasis.hh>
#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/functions/functionspacebases/subspacebasis.hh>
#include <dune/functions/functionspacebases/boundarydofs.hh>

#include <dune/functions/gridfunctions/discreteglobalbasisfunction.hh>
#include <dune/functions/gridfunctions/gridviewfunction.hh>

#include <dune/assembler/utility.hh>
#include <dune/assembler/defaultglobalassembler.hh>
#include <dune/assembler/denselocalassemblerbase.hh>
#include <dune/assembler/backends/istlvectorbackend.hh>
#include <dune/assembler/backends/istlmatrixbackend.hh>

#define BLOCKEDBASIS 1



template <class Basis>
auto sortedIndices (const Basis& basis)
{
  using MultiIndex = typename Basis::LocalView::MultiIndex;
  std::vector<MultiIndex> multiIndices;
  {
    auto localView = basis.localView();
    for (const auto& element : Dune::elements(basis.gridView()))
    {
      localView.bind(element);
      for (auto i : Dune::range(localView.size()))
        multiIndices.push_back(localView.index(i));
    }
    std::sort(multiIndices.begin(), multiIndices.end());
    auto lastUnique = std::unique(multiIndices.begin(), multiIndices.end());
    multiIndices.erase(lastUnique, multiIndices.end());
  }
  assert(multiIndices.size()==basis.dimension());
  return multiIndices;
}



template <class TestLocalView, class TrialLocalView, class F>
class LocalAssemblerWrapper
  : public Dune::Assembler::DenseLocalAssemblerBase<TestLocalView, TrialLocalView>
{
  using GridView = typename TestLocalView::GridView;
  using Element = typename TestLocalView::Element;

public:

  LocalAssemblerWrapper (const F& f)
    : f_(f)
  {}

  void bindLocalViews (const TestLocalView& testLocalView, const TrialLocalView& trialLocalView)
  {
    testLocalView_ = &testLocalView;
    trialLocalView_ = &trialLocalView;
  }

  void bindElement (const Element& element)
  {}

  template <class LocalMatrix>
  void assembleElementMatrix (const Element& e, LocalMatrix& localMatrix)
  {
    f_(localMatrix, *testLocalView_, *trialLocalView_);
  }

private:
  const TestLocalView* testLocalView_ = nullptr;
  const TrialLocalView* trialLocalView_ = nullptr;
  F f_;
};

template <class TestLocalView, class TrialLocalView, class F>
auto wrapLocalAssembler (const F& f)
{
  return LocalAssemblerWrapper<TestLocalView, TrialLocalView, F>(f);
}



template <class LocalPattern, class TestNode, class TrialNode>
void addNodeIndices (LocalPattern& localPattern, const TestNode& testNode, const TrialNode& trialNode)
{
  for (auto i: Dune::Assembler::Impl::localIndices(testNode))
    for (auto j: Dune::Assembler::Impl::localIndices(trialNode))
      localPattern.add(i, j);
}



template <class TestLocalView, class TrialLocalView, class TestTreePath, class TrialTreePath>
class LocalNodeAssembler
{
public:

  using GridView = typename TestLocalView::GridView;
  using Element = typename TestLocalView::Element;
  using TestNode = typename Dune::TypeTree::ChildForTreePath<typename TestLocalView::Tree, TestTreePath>;
  using TrialNode = typename Dune::TypeTree::ChildForTreePath<typename TrialLocalView::Tree, TrialTreePath>;

  LocalNodeAssembler (TestTreePath testTreePath, TrialTreePath trialTreePath)
    : testTreePath_(testTreePath)
    , trialTreePath_(trialTreePath)
  {}

  void bindLocalViews (const TestLocalView& testLocalView, const TrialLocalView& trialLocalView)
  {
    testLocalView_ = &testLocalView;
    trialLocalView_ = &trialLocalView;
    testNode_ = &Dune::TypeTree::child(testLocalView_->tree(), testTreePath_);
    trialNode_ = &Dune::TypeTree::child(trialLocalView_->tree(), trialTreePath_);
  }

  void bindElement (const Element& element)
  {}

  template <class LocalPattern>
  void assembleElementMatrixPattern (const Element& e, LocalPattern& localPattern) const
  {
    addNodeIndices(localPattern, *testNode_, *trialNode_);
  }

protected:
  const TestTreePath testTreePath_;
  const TrialTreePath trialTreePath_;
  const TestLocalView* testLocalView_ = nullptr;
  const TrialLocalView* trialLocalView_ = nullptr;
  const TestNode* testNode_ = nullptr;
  const TrialNode* trialNode_ = nullptr;
};



template <class LocalView, class TreePath>
class LocalVectorLaplaceAssembler
  : public LocalNodeAssembler<LocalView, LocalView, TreePath, TreePath>
{
  using Base = LocalNodeAssembler<LocalView, LocalView, TreePath, TreePath>;
  using Element = typename Base::Element;

  static constexpr int dim = Element::dimension;

  using Base::testLocalView_;
  using Base::trialLocalView_;
  using Base::testNode_;
  using Base::trialNode_;

public:

  LocalVectorLaplaceAssembler (TreePath treePath)
    : Base(treePath, treePath)
  {}

  template <class LocalMatrix>
  void assembleElementMatrix (const Element& element, LocalMatrix& localMatrix)
  {
    assert(testNode_ == trialNode_);

    auto geometry = element.geometry();

    using namespace Dune::Indices;

    const auto& finiteElement = testNode_->child(_0).finiteElement();
    referenceJacobians_.resize(finiteElement.size());
    jacobians_.resize(finiteElement.size());

    // Get a quadrature rule
    int order = 2*finiteElement.localBasis().order();
    const auto& quad = Dune::QuadratureRules<double, dim>::rule(element.type(), order);

    // Loop over all quadrature points
    for (const auto& quadPoint : quad)
    {
      const auto integrationWeigth = geometry.integrationElement(quadPoint.position()) * quadPoint.weight();

      const auto jacobianInverse = geometry.jacobianInverse(quadPoint.position());
      finiteElement.localBasis().evaluateJacobian(quadPoint.position(), referenceJacobians_);
      for (auto i : Dune::range(finiteElement.size()))
        jacobians_[i] = referenceJacobians_[i] * jacobianInverse;

      for (auto i : Dune::range(finiteElement.size()))
        for (auto j : Dune::range(finiteElement.size()))
          for (auto k : Dune::range(dim))
          {
            size_t row = testNode_->child(k).localIndex(i);
            size_t col = trialNode_->child(k).localIndex(j);
            localMatrix[row][col] += (jacobians_[i] * transpose(jacobians_[j])) * integrationWeigth;
          }
    }
  }

  template <class LocalPattern>
  void assembleElementMatrixPattern (const Element& e, LocalPattern& localPattern) const
  {
    for (auto k : Dune::range(dim))
      addNodeIndices(localPattern, testNode_->child(k), trialNode_->child(k));
  }

private:
  std::vector<Dune::FieldMatrix<double,1,dim>> referenceJacobians_;
  std::vector<Dune::FieldMatrix<double,1,dim>> jacobians_;
};

template <class LocalView, class TreePath>
auto localVectorLaplaceAssembler (const TreePath& treePath)
{
  return LocalVectorLaplaceAssembler<LocalView, TreePath>(treePath);
}



template <class TestLocalView, class TrialLocalView, class TestTreePath, class TrialTreePath, bool adjoint=false>
class LocalDivAssembler
  : public LocalNodeAssembler<TestLocalView, TrialLocalView, TestTreePath, TrialTreePath>
{
  using Base = LocalNodeAssembler<TestLocalView, TrialLocalView, TestTreePath, TrialTreePath>;
  using Element = typename Base::Element;

  static const int dim = Element::dimension;

  using Base::testLocalView_;
  using Base::trialLocalView_;
  using Base::testNode_;
  using Base::trialNode_;

public:

  using Base::Base;

  template <class LocalMatrix>
  void assembleElementMatrix (const Element& element, LocalMatrix& localMatrix)
  {
    auto geometry = element.geometry();

    using namespace Dune::Indices;

    auto assembleDiv = [&](const auto& velocityNode, const auto& pressureNode) {

      const auto& velocityLocalFiniteElement = velocityNode.child(0).finiteElement();
      const auto& pressureLocalFiniteElement = pressureNode.finiteElement();

      referenceJacobians_.resize(velocityLocalFiniteElement.size());
      jacobians_.resize(velocityLocalFiniteElement.size());
      pressureValues_.resize(pressureLocalFiniteElement.size());

      int order = velocityLocalFiniteElement.localBasis().order() * pressureLocalFiniteElement.localBasis().order();
      const auto& quad = Dune::QuadratureRules<double, dim>::rule(element.type(), order);

      for (const auto& quadPoint : quad)
      {
        const auto integrationWeigth = geometry.integrationElement(quadPoint.position()) * quadPoint.weight();

        const auto jacobianInverse = geometry.jacobianInverse(quadPoint.position());
        velocityLocalFiniteElement.localBasis().evaluateJacobian(quadPoint.position(), referenceJacobians_);
        for (auto i : Dune::range(velocityLocalFiniteElement.size()))
          jacobians_[i] = referenceJacobians_[i] * jacobianInverse;

        pressureLocalFiniteElement.localBasis().evaluateFunction(quadPoint.position(), pressureValues_);

        for (auto i : Dune::range(velocityLocalFiniteElement.size()))
          for (auto j : Dune::range(pressureLocalFiniteElement.size()))
            for (size_t k=0; k<dim; k++)
            {
              size_t vIndex = velocityNode.child(k).localIndex(i);
              size_t pIndex = pressureNode.localIndex(j);
              if constexpr (not adjoint)
                localMatrix[vIndex][pIndex] += jacobians_[i][0][k] * pressureValues_[j] * integrationWeigth;
              else
                localMatrix[pIndex][vIndex] += jacobians_[i][0][k] * pressureValues_[j] * integrationWeigth;
            }
      }
    };
    if constexpr (not adjoint)
      assembleDiv(*testNode_, *trialNode_);
    else
      assembleDiv(*trialNode_, *testNode_);
  }

private:
  std::vector<Dune::FieldMatrix<double,1,dim>> referenceJacobians_;
  std::vector<Dune::FieldMatrix<double,1,dim>> jacobians_;
  std::vector<Dune::FieldVector<double,1>> pressureValues_;
};


template <class TestLocalView, class TrialLocalView, class TestTreePath, class TrialTreePath>
auto localDivAssembler (const TestTreePath& testTreePath, const TrialTreePath& trialTreePath)
{
  return LocalDivAssembler<TestLocalView, TrialLocalView, TestTreePath, TrialTreePath>(testTreePath, trialTreePath);
}

template <class TestLocalView, class TrialLocalView, class TestTreePath, class TrialTreePath>
auto localAdjointDivAssembler (const TestTreePath& testTreePath, const TrialTreePath& trialTreePath)
{
  return LocalDivAssembler<TestLocalView, TrialLocalView, TestTreePath, TrialTreePath, true>(testTreePath, trialTreePath);
}

template <class TestLocalView, class TrialLocalView, class ... LocalAssemblers>
class LocalSumAssembler
{
  using GridView = typename TestLocalView::GridView;
  using Element = typename TestLocalView::Element;

  template <class LA>
  static constexpr bool isLocalElementMatrixAssembler =
    Dune::Assembler::Concepts::LocalElementMatrixAssembler<LA,TestLocalView,TrialLocalView>;

public:
  LocalSumAssembler (LocalAssemblers... localAssemblers)
    : localAssemblers_(localAssemblers ...)
  {}

  void bindLocalViews (const TestLocalView& testLocalView, const TrialLocalView& trialLocalView)
  {
    Dune::Hybrid::forEach(localAssemblers_, [&](auto&& localAssembler) {
      localAssembler.bindLocalViews(testLocalView, trialLocalView);
    });
  }

  void bindElement (const Element& element)
  {
    Dune::Hybrid::forEach(localAssemblers_, [&](auto&& localAssembler) {
      localAssembler.bindElement(element);
    });
  }

  template <class LocalPattern>
  requires(isLocalElementMatrixAssembler<LocalAssemblers> || ...)
  void assembleElementMatrixPattern (const Element& element, LocalPattern& localPattern)
  {
    Dune::Hybrid::forEach(localAssemblers_, [&](auto&& localAssembler) {
      using LocalAssembler = std::decay_t<decltype(localAssembler)>;
      if constexpr (isLocalElementMatrixAssembler<LocalAssembler>)
        localAssembler.assembleElementMatrixPattern(element, localPattern);
    });
  }

  template <class LocalMatrix>
  requires(isLocalElementMatrixAssembler<LocalAssemblers> || ...)
  void assembleElementMatrix (const Element& element, LocalMatrix& localMatrix)
  {
    Dune::Hybrid::forEach(localAssemblers_, [&](auto&& localAssembler) {
      using LocalAssembler = std::decay_t<decltype(localAssembler)>;
      if constexpr (isLocalElementMatrixAssembler<LocalAssembler>)
        localAssembler.assembleElementMatrix(element, localMatrix);
    });
  }

private:
  std::tuple<LocalAssemblers...> localAssemblers_;
};

template <class TestLocalView, class TrialLocalView, class ... LocalAssemblers>
auto localSumAssembler (LocalAssemblers... localAssemblers)
{
  return LocalSumAssembler<TestLocalView, TrialLocalView, std::decay_t<LocalAssemblers>...>(localAssemblers ...);
}



template <class LocalView, class TrialLocalView>
class LocalStokesAssembler
  : public Dune::Assembler::DenseLocalAssemblerBase<LocalView, TrialLocalView>
{
  using GridView = typename LocalView::GridView;
  using Element = typename LocalView::Element;

public:

  void bindLocalViews (const LocalView& testLocalView, const LocalView& trialLocalView)
  {
    testLocalView_ = &testLocalView;
    trialLocalView_ = &trialLocalView;
  }

  void bindElement (const Element& element)
  {}

  template <class LocalPattern>
  void assembleElementMatrixPattern (const Element& e, LocalPattern& localPattern) const
  {
    using namespace Dune::Indices;
    const auto& velocityNode = testLocalView_->tree().child(_0);
    const auto& pressureNode = testLocalView_->tree().child(_1);
    for (auto k: Dune::range(velocityNode.degree()))
      addNodeIndices(localPattern, velocityNode.child(k), velocityNode.child(k));
    addNodeIndices(localPattern, velocityNode, pressureNode);
    addNodeIndices(localPattern, pressureNode, velocityNode);
  }

  template <class LocalMatrix>
  void assembleElementMatrix (const Element& element, LocalMatrix& localMatrix) const
  {
    const auto dim = LocalView::Element::dimension;
    auto geometry = element.geometry();

    using namespace Dune::Indices;

    const auto& velocityNode = testLocalView_->tree().child(_0);
    const auto& pressureNode = testLocalView_->tree().child(_1);

    const auto& velocityLocalFiniteElement = velocityNode.child(0).finiteElement();
    const auto& pressureLocalFiniteElement = pressureNode.finiteElement();

    // Get a quadrature rule
    int order = 2*(dim*velocityLocalFiniteElement.localBasis().order()-1);
    const auto& quad = Dune::QuadratureRules<double, dim>::rule(element.type(), order);

    // Loop over all quadrature points
    for (const auto& quadPoint : quad)
    {
      // The inverse Jacobian of the map from the
      // reference element to the element
      const auto jacobianInverse = geometry.jacobianInverse(quadPoint.position());

      // The multiplicative factor in the integral transformation formula
      const auto integrationWeigth = geometry.integrationElement(quadPoint.position()) * quadPoint.weight();

      ///////////////////////////////////////////////////////////////////////
      //  Velocity--velocity coupling
      ///////////////////////////////////////////////////////////////////////

      // The gradients of the shape functions on the reference element
      std::vector<Dune::FieldMatrix<double,1,dim>> referenceJacobians;
      velocityLocalFiniteElement.localBasis().evaluateJacobian(quadPoint.position(), referenceJacobians);

      // Compute the shape function gradients on the grid element
      std::vector<Dune::FieldMatrix<double,1,dim>> jacobians(referenceJacobians.size());
      for (size_t i=0; i<jacobians.size(); i++)
        jacobians[i] = referenceJacobians[i] * jacobianInverse;

      // Compute the actual matrix entries
      for (auto i : Dune::range(velocityNode.child(0).size()))
        for (auto j : Dune::range(velocityNode.child(0).size()))
          for (auto k : Dune::range(dim))
          {
            size_t row = velocityNode.child(k).localIndex(i);
            size_t col = velocityNode.child(k).localIndex(j);
            localMatrix[row][col] += (jacobians[i] * transpose(jacobians[j])) * integrationWeigth;
          }

      ///////////////////////////////////////////////////////////////////////
      //  Velocity--pressure coupling
      ///////////////////////////////////////////////////////////////////////

      // The values of the pressure shape functions
      std::vector<Dune::FieldVector<double,1>> pressureValues;
      pressureLocalFiniteElement.localBasis().evaluateFunction(quadPoint.position(), pressureValues);

      // Compute the actual matrix entries
      for (auto i : Dune::range(velocityNode.child(0).size()))
        for (auto j : Dune::range(pressureNode.size()))
          for (size_t k=0; k<dim; k++)
          {
            size_t vIndex = velocityNode.child(k).localIndex(i);
            size_t pIndex = pressureNode.localIndex(j);
            localMatrix[vIndex][pIndex] += jacobians[i][0][k] * pressureValues[j] * integrationWeigth;
            localMatrix[pIndex][vIndex] += jacobians[i][0][k] * pressureValues[j] * integrationWeigth;
          }
    }
  }

private:
  const LocalView* testLocalView_ = nullptr;
  const LocalView* trialLocalView_ = nullptr;
};


using namespace Dune;


int main (int argc, char *argv[]) try
{
  // Set up MPI, if available
  MPIHelper::instance(argc, argv);

  std::size_t refinements = 0;
  if (argc>1)
    refinements = std::stoul(std::string(argv[1]));

  ///////////////////////////////////
  //   Generate the grid
  ///////////////////////////////////

  const int dim = 2;
  using GridType = YaspGrid<dim>;
  Dune::FieldVector<double,dim> upperRight = {1, 1};
  std::array<int,dim> elements = {{4, 4}};
  GridType grid(upperRight,elements);

  grid.globalRefine(refinements);

  using GridView = typename GridType::LeafGridView;
  GridView gridView = grid.leafGridView();

  /////////////////////////////////////////////////////////
  //   Choose a finite element space
  /////////////////////////////////////////////////////////

  using namespace Functions::BasisFactory;

  constexpr std::size_t p = 1; // pressure order for Taylor-Hood

  /////////////////////////////////////////////////////////
  // Define basis and suitable matrix and vector types
  /////////////////////////////////////////////////////////

#if BLOCKEDBASIS

  auto taylorHoodBasis = makeBasis(
    gridView,
    composite(
      power<dim>(
        lagrange<p+1>(),
        blockedInterleaved()),
      lagrange<p>(),
      blockedLexicographic()
      ));

  using VelocityVector = BlockVector<FieldVector<double,dim>>;
  using PressureVector = BlockVector<double>;
  using VectorType = MultiTypeBlockVector<VelocityVector, PressureVector>;

  using VelocityBitVector = BlockVector<FieldVector<char,dim>>;
  using PressureBitVector = BlockVector<char>;
  using BitVectorType = MultiTypeBlockVector<VelocityBitVector, PressureBitVector>;

  //  using Matrix00 = BCRSMatrix<FieldMatrix<double,dim,dim>>;
  using Matrix00 = BCRSMatrix<DiagonalMatrix<double,dim>>;
  // Use Single(Row|Column)Matrix to tag matrices that do not consume
  // a row of column index. This is needed to securely handle hybrid indices.
  using Matrix01 = BCRSMatrix<Dune::Assembler::SingleColumnMatrix<FieldMatrix<double,dim,1>>>;
  using Matrix10 = BCRSMatrix<Dune::Assembler::SingleRowMatrix<FieldMatrix<double,1,dim>>>;
  using Matrix11 = BCRSMatrix<double>;
  using MatrixRow0 = MultiTypeBlockVector<Matrix00, Matrix01>;
  using MatrixRow1 = MultiTypeBlockVector<Matrix10, Matrix11>;
  using MatrixType = MultiTypeBlockMatrix<MatrixRow0,MatrixRow1>;

#else

  auto taylorHoodBasis = makeBasis(
    gridView,
    composite(
      power<dim>(
        lagrange<p+1>(),
        flatInterleaved()),
      lagrange<p>()
      ));

  using VectorType = BlockVector<BlockVector<double>>;
  using BitVectorType = BlockVector<BlockVector<double>>;
  using MatrixType = Matrix<BCRSMatrix<double>>;
#endif

  std::cout << "Dimension of discrete space is " << taylorHoodBasis.dimension() << std::endl;

  using Basis = decltype(taylorHoodBasis);


  /////////////////////////////////////////////////////////
  //  Assemble the system
  /////////////////////////////////////////////////////////

  VectorType rhs;

  auto rhsBackend = Dune::Assembler::ISTLVectorBackend(rhs);

  rhsBackend.resize(taylorHoodBasis);
  rhsBackend.setZero();

  MatrixType stiffnessMatrix;

  // In dune-fufem one could alternatively use:
  //
  // using namespace Dune::Fufem::Forms;
  // auto velocityBasis = Functions::subspaceBasis(taylorHoodBasis, _0);
  // auto pressureBasis = Functions::subspaceBasis(taylorHoodBasis, _1);
  // auto u = trialFunction(velocityBasis);
  // auto p = trialFunction(pressureBasis);
  // auto q = testFunction(pressureBasis);
  // auto v = testFunction(velocityBasis);
  // auto localAssembler = integrate( dot(grad(u), grad(v)) + div(u)*q + div(v)*p );
  //
  // Then assembling the entries would be about 33% faster compared to
  // the handwritten assembler above and take essentially as much time
  // as filling the pattern or exporting it to a BCRSMatrix (measured
  // on a larger 1000x1000 grid with 4 threads).

  // Create dune-fufem global assembler
  auto assembler = Dune::Assembler::Assembler<Basis>(taylorHoodBasis);

  using LocalView = Basis::LocalView;

  using Dune::TypeTree::treePath;
  using namespace Dune::Indices;

  auto&& matrixBackend = Dune::Assembler::ISTLMatrixBackend(stiffnessMatrix);
  auto patternBuilder = matrixBackend.patternBuilder();

  auto timer = Dune::Timer();

#if 0
  // Variant A: Assemble with all-at-once Stokes assembler
  {
    auto localAssembler = LocalStokesAssembler<LocalView, LocalView>();

    patternBuilder.resize(taylorHoodBasis, taylorHoodBasis);

    assembler.assembleMatrixPattern(localAssembler, patternBuilder);

    patternBuilder.setupMatrix();

    assembler.assembleMatrixEntries(localAssembler, matrixBackend);
  }
#elif 0
  // Variant B: Assemble by calling separate assemblers globally
  {
    auto laplaceAssembler = localVectorLaplaceAssembler<LocalView>(treePath(_0));
    auto divAssembler = localDivAssembler<LocalView, LocalView>(treePath(_0), treePath(_1));
    auto adjointDivAssembler = localAdjointDivAssembler<LocalView, LocalView>(treePath(_1), treePath(_0));

    patternBuilder.resize(taylorHoodBasis, taylorHoodBasis);

    assembler.assembleMatrixPattern(laplaceAssembler, patternBuilder);
    assembler.assembleMatrixPattern(divAssembler, patternBuilder);
    assembler.assembleMatrixPattern(adjointDivAssembler, patternBuilder);

    patternBuilder.setupMatrix();

    assembler.assembleMatrixEntries(laplaceAssembler, matrixBackend);
    assembler.assembleMatrixEntries(divAssembler, matrixBackend);
    assembler.assembleMatrixEntries(adjointDivAssembler, matrixBackend);
  }
#else
  // Variant C: Assemble by using a sum of separate local assemblers
  {
    auto laplaceAssembler = localVectorLaplaceAssembler<LocalView>(treePath(_0));
    auto divAssembler = localDivAssembler<LocalView, LocalView>(treePath(_0), treePath(_1));
    auto adjointDivAssembler = localAdjointDivAssembler<LocalView, LocalView>(treePath(_1), treePath(_0));

    auto localAssembler = localSumAssembler<LocalView, LocalView>(laplaceAssembler, divAssembler, adjointDivAssembler);

    patternBuilder.resize(taylorHoodBasis, taylorHoodBasis);

    assembler.assembleMatrixPattern(localAssembler, patternBuilder);

    patternBuilder.setupMatrix();

    assembler.assembleMatrixEntries(localAssembler, matrixBackend);
  }
#endif

  std::cout << "Assembling the problem took " << timer.elapsed() << "s" << std::endl;


  /////////////////////////////////////////////////////////
  // Set Dirichlet values.
  // Only velocity components have Dirichlet boundary values
  /////////////////////////////////////////////////////////

  BitVectorType isBoundary;

  auto isBoundaryBackend = Dune::Assembler::ISTLVectorBackend(isBoundary);
  isBoundaryBackend.resize(taylorHoodBasis);
  isBoundaryBackend.setZero();

  using namespace Indices;
  Functions::forEachBoundaryDOF(
    Functions::subspaceBasis(taylorHoodBasis, _0),
    [&] (auto&& index) {
    isBoundaryBackend[index] = true;
  });

  using Coordinate = GridView::Codim<0> ::Geometry::GlobalCoordinate;
  using VelocityRange = FieldVector<double,dim>;
  auto&& velocityDirichletData = [](Coordinate x)
  {
    return VelocityRange{0.0, double(x[0] < 1e-8)};
  };

  Functions::interpolate(
    Functions::subspaceBasis(taylorHoodBasis, _0), rhs,
    velocityDirichletData,
    isBoundary);

  ////////////////////////////////////////////
  //   Modify Dirichlet rows
  ////////////////////////////////////////////

  // Dirty hack: First create flat-to-multiindex map
  // and then use flatMatrixForEach to traverse.
  // If we would have matrixForEach() which uses
  // multi-indices, this would look nicer, but maybe be
  // less efficient.
  auto multiIndexFor = sortedIndices(taylorHoodBasis);
  Dune::flatMatrixForEach(stiffnessMatrix, [&](auto& entry, auto row, auto col) {
    if (isBoundaryBackend[multiIndexFor[row]])
      entry = (row==col) ? 1 : 0;
  });

  ////////////////////////////
  //   Compute solution
  ////////////////////////////
  // Start from the rhs vector; that way the Dirichlet entries are already correct
  VectorType x = rhs;

  // Technicality:  turn the matrix into a linear operator
  MatrixAdapter<MatrixType,VectorType,VectorType> stiffnessOperator(stiffnessMatrix);

  // Fancy (but only) way to not have a preconditioner at all
  Richardson<VectorType,VectorType> preconditioner(1.0);

  // Construct the actual iterative solver
  RestartedGMResSolver<VectorType> solver(
    stiffnessOperator,        // operator to invert
    preconditioner,           // preconditioner for iteration
    1e-10,                    // desired residual reduction factor
    500,                      // number of iterations between restarts
    500,                      // maximum number of iterations
    2);                       // verbosity of the solver

  // Object storing some statistics about the solving process
  InverseOperatorResult statistics;

  // Solve!
  solver.apply(x, rhs, statistics);

  ////////////////////////////////////////////////////////////////////////////
  //  Make a discrete function from the FE basis and the coefficient vector
  ////////////////////////////////////////////////////////////////////////////

  using VelocityRange = FieldVector<double,dim>;
  using PressureRange = double;

  auto velocityFunction
    = Functions::makeDiscreteGlobalBasisFunction<VelocityRange>(
        Functions::subspaceBasis(taylorHoodBasis, _0), x);
  auto pressureFunction
    = Functions::makeDiscreteGlobalBasisFunction<PressureRange>(
        Functions::subspaceBasis(taylorHoodBasis, _1), x);

  //////////////////////////////////////////////////////////////////////////////////////////////
  //  Write result to VTK file
  //  We need to subsample, because VTK cannot natively display real second-order functions
  //////////////////////////////////////////////////////////////////////////////////////////////
  SubsamplingVTKWriter<GridView> vtkWriter(
    gridView,
    refinementLevels(2));
  vtkWriter.addVertexData(
    velocityFunction,
    VTK::FieldInfo("velocity", VTK::FieldInfo::Type::vector, dim));
  vtkWriter.addVertexData(
    pressureFunction,
    VTK::FieldInfo("pressure", VTK::FieldInfo::Type::scalar, 1));
  vtkWriter.write("stokes-taylorhood-result");

}
// Error handling
catch (Exception& e) {
  std::cout << e.what() << std::endl;
}
